import logging
import urllib.request
import json
from typing import Literal, Optional, Type, Union, Annotated
from pydantic import (
    BaseModel,
    Field,
    ConfigDict,
    field_validator,
    AliasChoices,
    Discriminator,
    FilePath,
    PrivateAttr,
)
import datetime
from typing import Dict, Any
from joserfc.jwk import KeySet, RSAKey, OctKey, ECKey
from joserfc import jwt
from joserfc.jwk import GuestProtocol
from authentikate.errors import JwksError, MalformedJwtTokenError


logger = logging.getLogger(__name__)


class JWTToken(BaseModel):
    """A JWT token

    This is a pydantic model that represents a JWT token.
    It is used to validate the token and to extract information from it.
    The token is decoded using the `decode_token` function.

    """

    model_config = ConfigDict(extra="ignore")

    sub: str
    """A unique identifier for the user (is unique for the issuer)"""
    iss: str
    """The issuer of the token"""

    exp: datetime.datetime
    """The expiration time of the token"""

    active_org: str | None = None
    """The active organization of the user, if any"""

    client_id: str
    """The client_id of the app that requested the token"""
    preferred_username: str
    """The username of the user"""
    roles: list[str]
    """The roles of the user"""
    scope: str
    """The scope of the token"""

    iat: datetime.datetime
    """The issued at time of the token"""

    aud: list[str] | None = None
    """The audience of the token"""

    jti: str | None = None
    """The unique identifier for the token"""

    raw: str
    """ The raw original token string """

    client_app: str | None = None
    """ The client app name """

    client_release: str | None = None
    """ The client release version """

    client_device: str | None = None
    """ The client device identifier """

    @field_validator("aud", mode="before")
    def aud_to_list(
        cls: Type["JWTToken"], v: str | list[str] | None
    ) -> list[str] | None:
        """Convert the aud to a list"""
        if not v:
            return None
        if isinstance(v, str):
            return [v]

        return v

    @field_validator("sub", mode="before")
    def sub_to_username(cls: Type["JWTToken"], v: str) -> str:
        """Convert the sub to a username compatible string"""
        if isinstance(v, int):
            return str(v)
        return v

    @field_validator("iat", mode="before")
    def iat_to_datetime(cls: Type["JWTToken"], v: int) -> datetime.datetime:
        """Convert the iat to a datetime object"""
        if v is None:
            return None
        if isinstance(v, int):
            return datetime.datetime.fromtimestamp(v)
        return v

    @field_validator("exp", mode="before")
    def exp_to_datetime(cls: Type["JWTToken"], v: int) -> datetime.datetime:
        """Convert the exp to a datetime object"""
        if isinstance(v, int):
            return datetime.datetime.fromtimestamp(v)
        return v

    @property
    def changed_hash(self) -> str:
        """A hash that changes when the user changes"""
        return str(
            hash(
                self.sub
                + self.preferred_username
                + " ".join(self.roles)
                + (self.active_org or "")
            )
        )

    @property
    def scopes(self) -> list[str]:
        """The scopes of the token. Each scope is a string separated by a space"""
        return self.scope.split(" ")

    def has_scopes(self, scopes: list[str]) -> bool:
        """Check if the user has the given scope"""
        if not scopes:
            return True

        return all(scope in self.scopes for scope in scopes)

    def has_any_role(self, roles: list[str]) -> bool:
        """Check if the user has any of the given roles"""
        if not roles:
            return True

        return any(role in self.roles for role in roles)

    def has_roles(self, roles: list[str]) -> bool:
        """Check if the user has the given role"""
        if not roles:
            return True

        return all(role in self.roles for role in roles)

    def has_any_scope(self, scopes: list[str]) -> bool:
        """Check if the user has any of the given scopes"""
        if not scopes:
            return True

        return any(scope in self.scopes for scope in scopes)


class StaticToken(JWTToken):
    """A static JWT token"""

    sub: str
    iss: str = "static_issuer"
    """The issuer of the token"""
    iat: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now())
    exp: datetime.datetime = Field(
        default_factory=lambda: datetime.datetime.now() + datetime.timedelta(days=1)
    )
    client_id: str = "static"
    client_app: str = "static_app"
    client_release: str = "v1.0.0"
    client_device: str = "static_device"
    active_org: str = "static_org"
    preferred_username: str = "static_user"
    scope: str = "openid profile email"
    roles: list[str] = Field(default_factory=lambda: ["admin"])
    raw: str = Field(default_factory=lambda: "static_token")


class ImitationRequest(BaseModel):
    """An imitation request"""

    sub: str
    iss: str


class Issuer(BaseModel):
    """An issuer

    This is a pydantic model that represents an issuer.
    It is used to validate the issuer and to extract information from it.
    """

    model_config = ConfigDict(extra="forbid")
    kind: str
    iss: str = Field(
        validation_alias=AliasChoices("iss", "issuer", "issuer_url", "ISSUER")
    )
    """The issuer of the token"""

    def get_as_jwks(self) -> list[Dict[str, Any]]:
        """Get the jwks of the issuer"""
        raise NotImplementedError(
            "get_jwks not implemented. Must be implemented in subclass"
        )

    def refresh(self) -> None:
        """Refresh the issuer jwks if applicable"""
        pass


class JWKIssuer(Issuer):
    """An issuer

    This is a pydantic model that represents an issuer.
    It is used to validate the issuer and to extract information from it.
    """

    kind: Literal["jwks_dict"] = Field(
        default="jwks_dict",
    )

    iss: str = Field(
        validation_alias=AliasChoices("iss", "issuer", "issuer_url", "ISSUER")
    )
    """The issuer of the token"""

    jwks: Dict[str, Any] = Field(
        validation_alias=AliasChoices("jwks", "JWKS", "JWKS_DICT")
    )
    """The jwks of the issuer"""

    @field_validator("jwks", mode="before")
    def validate_jwks_dict(cls: Type["JWKIssuer"], v: Dict[str, Any]) -> Dict[str, Any]:
        """Validate the jwks dict"""
        if not isinstance(v, dict):
            raise ValueError("jwks_dict must be a dict")
        if "keys" not in v:
            raise ValueError("jwks_dict must contain a keys field")
        if not isinstance(v["keys"], list):
            raise ValueError("jwks_dict keys must be a list")
        return v

    def get_as_jwks(self) -> list[Dict[str, Any]]:
        """Get the jwks of the issuer"""
        return self.jwks["keys"]


class RSAKeyIssuer(Issuer):
    """An issuer

    This is a pydantic model that represents an issuer.
    It is used to validate the issuer and to extract information from it.
    """

    model_config = ConfigDict(extra="forbid")
    kind: Literal["rsa"] = Field(
        default="rsa",
    )

    iss: str = Field(
        validation_alias=AliasChoices("iss", "issuer", "issuer_url", "ISSUER")
    )
    key_id: str = Field(
        default="1", validation_alias=AliasChoices("key_id", "kid", "KID")
    )
    """The issuer of the token"""
    public_key: str = Field(validation_alias=AliasChoices("public_key", "PUBLIC_KEY"))

    def get_as_jwks(self) -> list[Dict[str, Any]]:
        """Get the jwks of the issuer"""
        t = RSAKey.import_key(self.public_key)
        return [t.as_dict(kid=self.key_id)]


class RSAKeyFileIssuer(Issuer):
    """An issuer

    This is a pydantic model that represents an issuer.
    It is used to validate the issuer and to extract information from it.
    """

    model_config = ConfigDict(extra="forbid")
    kind: Literal["rsa_file"] = Field(
        default="rsa_file",
    )

    iss: str = Field(
        validation_alias=AliasChoices("iss", "issuer", "issuer_url", "ISSUER")
    )
    key_id: str = Field(
        default="1", validation_alias=AliasChoices("key_id", "kid", "KID")
    )
    """The issuer of the token"""
    public_key_pem_file: FilePath = Field(
        validation_alias=AliasChoices("public_key_pem_file", "PUBLIC_KEY_PEM_FILE")
    )

    def get_as_jwks(self) -> list[Dict[str, Any]]:
        """Get the jwks of the issuer"""

        with open(self.public_key_pem_file, "rb") as f:
            public_key = f.read()

        t = RSAKey.import_key(public_key)
        return [t.as_dict(kid=self.key_id)]


class JWKSUriIssuer(Issuer):
    """An issuer

    This is a pydantic model that represents an issuer.
    It is used to validate the issuer and to extract information from it.
    """

    model_config = ConfigDict(extra="forbid")
    kind: Literal["jwks_uri"] = Field(
        default="jwks_uri",
    )

    iss: str = Field(
        validation_alias=AliasChoices("iss", "issuer", "issuer_url", "ISSUER")
    )
    """The issuer of the token"""
    jwks_uri: str = Field(validation_alias=AliasChoices("jwks_uri", "JWKS_URI"))
    _cache: list[Dict[str, Any]] | None = PrivateAttr(default=None)

    def get_as_jwks(self) -> list[Dict[str, Any]]:
        """Get the jwks of the issuer"""

        if self._cache is None:
            self.refresh()

        return self._cache  # type: ignore

    def refresh(self) -> None:
        """Refresh the jwks from the uri"""
        try:
            with urllib.request.urlopen(self.jwks_uri) as response:
                data = json.loads(response.read())
                self._cache = data["keys"]
        except Exception as e:
            raise JwksError(f"Error fetching jwks from {self.jwks_uri}") from e


IssuerUnion = Annotated[
    Union[JWKIssuer, RSAKeyIssuer, RSAKeyFileIssuer, JWKSUriIssuer], Discriminator("kind")
]



class AuthentikateSettings(BaseModel):
    """The settings for authentikate

    This is a pydantic model that represents the settings for authentikate.
    It is used to configure the library.
    """

    model_config = ConfigDict(extra="forbid")

    issuers: list[IssuerUnion] = Field(
        validation_alias=AliasChoices(
            "issuers",
            "iss",
            "issuer",
            "issuer_url",
            "ISSUERS",
        )
    )
    authorization_headers: list[str] = Field(
        default_factory=lambda: [
            "Authorization",
            "X-Authorization",
            "AUTHORIZATION",
            "authorization",
        ],
        validation_alias=AliasChoices(
            "authorization_headers", "AUTHORIZATION_HEADERS", "AUTHORIZATION_HEADERS"
        ),
    )
    static_tokens: dict[str, StaticToken] = Field(
        default_factory=dict,
        validation_alias=AliasChoices(
            "static_tokens", "STATIC_TOKENS", "STATIC_TOKENS"
        ),
    )
    """A map of static tokens to their decoded values. Should only be used in tests."""

    def get_jwks(self) -> list[Dict[str, Any]]:
        """Get the jwks of the issuer"""

        merged_jwks = {}

        for issuer in self.issuers:
            keys = issuer.get_as_jwks()

            if not isinstance(keys, list):
                raise JwksError("keys must be a list")

            for key in keys:
                if key.get("kid") is None:
                    raise JwksError("key must contain a kid field")

                if key["kid"] in merged_jwks:
                    raise JwksError(f"Duplicate kid found: {key['kid']}")

                merged_jwks[key["kid"]] = key

        if not merged_jwks:
            raise JwksError("No keys found in jwks")

        validated_keys = []

        for key in merged_jwks.values():
            validated_keys.append(key)

        return validated_keys

    def load_key(self, obj: GuestProtocol) -> KeySet:
        """Resolve the key from the header"""
        kid = obj.headers().get("kid")
        if not kid:
            raise MalformedJwtTokenError("Missing kid in header")

        jwks = self.get_jwks()

        # Check if the kid is in the jwks
        found = False
        for key in jwks:
            if key.get("kid") == kid:
                found = True
                break

        if not found:
            # Trigger refresh on all issuers
            for issuer in self.issuers:
                issuer.refresh()
            # Re-fetch
            jwks = self.get_jwks()

        key_set = KeySet.import_key_set({"keys": jwks})
        return key_set



from django.db import models  # Create your models here.
from django.contrib.auth.models import AbstractUser


class Organization(models.Model):
    """An Organization model to represent an organization in the system"""

    slug = models.CharField(max_length=1000, unique=True)

    def __str__(self) -> str:
        """String representation of Organization"""
        return self.slug


class User(AbstractUser):
    """A reflection on the real User"""

    sub = models.CharField(max_length=1000, null=True, blank=True)
    iss = models.CharField(max_length=1000, null=True, blank=True)
    active_organization = models.ForeignKey(
        Organization,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="active_users",
    )
    changed_hash = models.CharField(max_length=1000, null=True, blank=True)

    class Meta:
        """Meta class for User"""

        constraints = [
            models.UniqueConstraint(
                fields=["sub", "iss"],
                condition=models.Q(sub__isnull=False, iss__isnull=False),
                name="unique_sub_iss_if_both_not_null",
            )
        ]
        permissions = [("imitate", "Can imitate me")]


class Membership(models.Model):
    """A Membership model to represent a user's membership in an organization"""

    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name="memberships")
    organization = models.ForeignKey(
        Organization, on_delete=models.CASCADE, related_name="memberships"
    )
    blocked = models.BooleanField(default=False)
    roles = models.JSONField(default=list)

    class Meta:
        """Meta class for Membership"""

        unique_together = ("user", "organization")

    def __str__(self) -> str:
        """String representation of Membership"""
        return f"{self.user} in {self.organization}"


class Device(models.Model):
    """A Device model to represent a user's device in the system"""

    device_id = models.CharField(max_length=2000, unique=True)

    def __str__(self) -> str:
        """String representation of Device"""
        return f"{self.device_id}"


class App(models.Model):
    """An App model to represent an application in the system"""

    identifier = models.CharField(max_length=2000)

    def __str__(self) -> str:
        """String representation of App"""
        return f"{self.identifier}"


class Release(models.Model):
    """A Release model to represent a release of an application in the system"""

    app = models.ForeignKey(App, on_delete=models.CASCADE, related_name="releases")
    version = models.CharField(max_length=2000)

    class Meta:
        """Meta class for Release"""


class Client(models.Model):
    """An Oauth2 Client

    An Oauth2 Client is a model to represent an Oauth2 client that is
    registered when a JWT token is authenticated. It retrieves
    the client_id from the token and uses it to create a new
    app or retrieve an existing app. This allows for the grouping
    of users by app.

    """

    device = models.ForeignKey(Device, on_delete=models.SET_NULL, null=True, blank=True)
    release = models.ForeignKey(
        Release,
        on_delete=models.SET_NULL,
        related_name="clients",
        null=True,
        blank=True,
    )
    iss = models.CharField(max_length=2000, null=True, blank=True)
    client_id = models.CharField(max_length=2000)
    name = models.CharField(max_length=2000, null=True, blank=True)

    class Meta:
        """Meta class for Client"""

        unique_together = ("iss", "client_id")

    def __str__(self) -> str:
        """String representation of Client"""
        return f"{self.name}"



from typing import Any, AsyncIterator, cast
from strawberry.extensions import SchemaExtension
from kante.context import WsContext, HttpContext
from authentikate.vars import token_var, user_var, client_var, organization_var
from authentikate.base_models import AuthentikateSettings, JWTToken
from authentikate.models import Client, User
from authentikate.utils import authenticate_header, authenticate_token
from authentikate.protocols import UserModel, OrganizationModel, MembershipModel


class AuthentikateExtension(SchemaExtension):
    """This is the extension class for the authentikate extension"""

    def get_settings(self) -> "AuthentikateSettings":
        """Get the settings for the extension"""
        from authentikate.settings import get_settings

        # Call the function to get the settings
        settings = get_settings()
        return settings

    async def aexpand_user_from_token(self, token: JWTToken) -> User:
        """Expand a user from the provided JWT token"""
        from authentikate.expand import aexpand_user_from_token

        # Call the async function to expand the user
        user = await aexpand_user_from_token(token)
        return cast(User, user)

    async def aexpand_client_from_token(self, token: JWTToken) -> Client:
        """Expand a client from the provided JWT token"""
        from authentikate.expand import aexpand_client_from_token

        # Call the async function to expand the client
        client = await aexpand_client_from_token(token)
        return cast(Client, client)

    async def aexpand_organization_from_token(
        self, token: JWTToken
    ) -> "OrganizationModel":
        """Expand an organization from the provided JWT token"""
        from authentikate.expand import aexpand_organization_from_token

        # Call the async function to expand the organization
        organization = await aexpand_organization_from_token(token)
        return cast(OrganizationModel, organization)

    async def aexpand_membership_from_user_and_organization(
        self, user: UserModel, organization: OrganizationModel, token: JWTToken
    ) -> "MembershipModel":
        """Expand a membership from the provided JWT token"""
        from authentikate.expand import aexpand_membership

        # Call the async function to expand the membership
        membership = await aexpand_membership(user, organization, token)
        return cast(MembershipModel, membership)

    async def on_operation(self) -> AsyncIterator[None]:
        """Set the token in the context variable"""

        context = self.execution_context.context

        reset_user = None
        reset_client = None
        reset_token = None
        reset_organization = None

        if isinstance(context, WsContext):
            # WebSocket context
            # Do something with the WebSocket context

            token = authenticate_token(
                context.connection_params.get("token", ""),
                self.get_settings(),
            )
            reset_token = token_var.set(token)
            if token:
                user = await self.aexpand_user_from_token(token)
                client = await self.aexpand_client_from_token(token)
                organization = await self.aexpand_organization_from_token(token)

                membership = await self.aexpand_membership_from_user_and_organization(
                    cast(UserModel, user), organization, token
                )

                reset_client = client_var.set(client)
                reset_user = user_var.set(cast(UserModel, user))
                reset_organization = organization_var.set(organization)

                context.request.set_user(cast(Any, user))
                context.request.set_client(cast(Any, client))
                context.request.set_membership(membership)
                context.request.set_organization(organization)
                context.request.set_extension("token", token)

        elif isinstance(context, HttpContext):
            # HTTP context
            # Do something with the HTTP context
            token = authenticate_header(
                dict(context.headers),
                self.get_settings(),
            )
            reset_token = token_var.set(token)
            if token:
                user = await self.aexpand_user_from_token(token)
                client = await self.aexpand_client_from_token(token)
                organization = await self.aexpand_organization_from_token(token)

                membership = await self.aexpand_membership_from_user_and_organization(
                    cast(UserModel, user), organization, token
                )

                reset_client = client_var.set(client)
                reset_user = user_var.set(cast(UserModel, user))
                reset_organization = organization_var.set(organization)

                context.request.set_user(cast(Any, user))
                context.request.set_client(cast(Any, client))
                context.request.set_membership(membership)
                context.request.set_organization(organization)
                context.request.set_extension("token", token)
        else:
            raise ValueError(
                "Unknown context type. Cannot determine if it's WebSocket or HTTP."
            )

        yield

        # Cleanup
        if reset_user:
            user_var.reset(reset_user)

        if reset_client:
            client_var.reset(reset_client)

        if reset_token:
            token_var.reset(reset_token)

        if reset_organization:
            organization_var.reset(reset_organization)

        return



from django.contrib import admin
from authentikate.models import Client, User, Membership

# Register your models here.

admin.site.register(User)
admin.site.register(Client)
admin.site.register(Membership)



from typing import Protocol


class AppModel(Protocol):
    """A protocol for the app model"""

    id: int
    """The id of the app"""

    identifier: str
    """The identifier of the app"""


class ReleaseModel(Protocol):
    """A protocol for the release model"""

    id: int
    """The id of the release"""

    version: str
    """The version of the release"""

    app: AppModel
    """The app of the release"""


class UserModel(Protocol):
    """A protocol for the user model"""

    id: int
    """The id of the user"""


class OrganizationModel(Protocol):
    """A protocol for the organizaition model"""

    id: int
    """The id of the client"""

    slug: str
    """The name of orgnaization"""


class MembershipModel(Protocol):
    """A protocol for the membership model"""

    id: int
    """The id of the membership"""



# Generated by Django 5.2 on 2026-03-11 00:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("authentikate", "0004_device_client_device"),
    ]

    operations = [
        migrations.AlterField(
            model_name="client",
            name="client_id",
            field=models.CharField(max_length=2000),
        ),
    ]



from django.contrib.auth.models import Group
from authentikate import base_models, models
import logging
from authentikate.protocols import (
    UserModel,
    OrganizationModel,
)

logger = logging.getLogger(__name__)


async def aresolve_client_relations(
    token: base_models.JWTToken,
) -> tuple[models.Release | None, models.Device | None]:
    """Resolve related models referenced by a client token."""
    release = None
    device = None

    if token.client_app and token.client_release:
        app, _ = await models.App.objects.aget_or_create(identifier=token.client_app)
        release, _ = await models.Release.objects.aget_or_create(
            app=app, version=token.client_release
        )

    if token.client_device:
        device, _ = await models.Device.objects.aget_or_create(
            device_id=token.client_device
        )

    return release, device


def resolve_client_relations(
    token: base_models.JWTToken,
) -> tuple[models.Release | None, models.Device | None]:
    """Resolve related models referenced by a client token."""
    release = None
    device = None

    if token.client_app and token.client_release:
        app, _ = models.App.objects.get_or_create(identifier=token.client_app)
        release, _ = models.Release.objects.get_or_create(
            app=app, version=token.client_release
        )

    if token.client_device:
        device, _ = models.Device.objects.get_or_create(device_id=token.client_device)

    return release, device


def token_to_username(token: base_models.JWTToken) -> str:
    """Convert a JWT token to a username

    Parameters
    ----------
    token : structs.JWTToken
        The token to convert

    Returns
    -------

    str
        The username



    """
    # Generate a username based on the token's iss and sub
    # and ensure it's unique
    return f"{token.iss}_{token.sub}"


async def aset_user_groups(user: models.User, roles: list[str]) -> None:
    """Add a list of roles to a user

    Roles are added as groups

    Parameters
    ----------
    user : models.User
        The user to add the roles to
    roles : list[str]
        The roles to add
    """
    for role in roles:
        g, _ = await Group.objects.aget_or_create(name=role)
        await user.groups.aadd(g)


def set_user_groups(user: models.User, roles: list[str]) -> None:
    """Add a list of roles to a user

    Roles are added as groups

    Parameters
    ----------
    user : models.User
        The user to add the roles to
    roles : list[str]
        The roles to add
    """
    for role in roles:
        g, _ = Group.objects.get_or_create(name=role)
        user.groups.add(g)


async def aexpand_organization_from_token(
    token: base_models.JWTToken,
) -> models.Organization:
    """
    Expand an organization from the provided JWT token.
    """
    org, _ = await models.Organization.objects.aget_or_create(slug=token.active_org)
    return org


async def aexpand_membership(
    user: UserModel, organization: OrganizationModel, token: base_models.JWTToken
) -> models.Membership:
    """
    Expand a membership from the provided user and organization.


    """
    membership, _ = await models.Membership.objects.aupdate_or_create(
        user_id=user.id,
        organization_id=organization.id,
        defaults=dict(
            roles=token.roles,
        ),
    )
    assert membership.blocked is False, "Membership is blocked"
    return membership


async def aexpand_user_from_token(
    token: base_models.JWTToken,
) -> models.User:
    """
    Expand a user from the provided JWT token.
    """

    try:
        user = await models.User.objects.aget(sub=token.sub, iss=token.iss)
        if user.changed_hash != token.changed_hash:
            # User has changed, update the user object
            user.first_name = token.preferred_username
            user.changed_hash = token.changed_hash

            if token.active_org:
                current_org, _ = await models.Organization.objects.aget_or_create(
                    slug=token.active_org or "",
                )

                user.active_organization = current_org

            await user.asave()
            await aset_user_groups(user, token.roles)

        return user

    except models.User.DoesNotExist:

        user = models.User(
            sub=token.sub,
            username=token_to_username(token),
            iss=token.iss,
        )
        user.set_unusable_password()
        user.first_name = token.preferred_username
        user.changed_hash = token.changed_hash

        if token.active_org:
            current_org, _ = await models.Organization.objects.aget_or_create(
                slug=token.active_org or "",
            )

            user.active_organization = current_org

        await user.asave()
        return user


def expand_user_from_token(
    token: base_models.JWTToken,
) -> models.User:
    """
    Expand a user from the provided JWT token.
    """

    try:
        user = models.User.objects.get(sub=token.sub, iss=token.iss)
        if user.changed_hash != token.changed_hash:
            # User has changed, update the user object
            user.first_name = token.preferred_username
            user.changed_hash = token.changed_hash
            set_user_groups(user, token.roles)

            if token.active_org:
                current_org, _ = models.Organization.objects.get_or_create(
                    identifier=token.active_org
                )

                user.active_organization = current_org

            user.save()

        return user

    except models.User.DoesNotExist:

        user = models.User(
            sub=token.sub,
            username=(token_to_username(token)),
            iss=token.iss,
            first_name=token.preferred_username,
        )
        user.set_unusable_password()
        user.first_name = token.preferred_username
        user.changed_hash = token.changed_hash

        if token.active_org:
            current_org, _ = models.Organization.objects.get_or_create(
                identifier=token.active_org
            )

            user.active_organization = current_org

        user.save()
        set_user_groups(user, token.roles)
        return user


async def aexpand_client_from_token(
    token: base_models.JWTToken,
) -> models.Client:
    """
    Expand a client from the provided JWT token.
    """
    release, device = await aresolve_client_relations(token)
    client, _ = await models.Client.objects.aget_or_create(
        client_id=token.client_id,
        iss=token.iss,
        defaults={"release": release, "device": device},
    )

    if getattr(client, "device_id", None) is None and device:
        client.device = device
        await client.asave(update_fields=["device"])

    if getattr(client, "release_id", None) is None and release:
        client.release = release
        await client.asave(update_fields=["release"])

    return client


def expand_client_from_token(
    token: base_models.JWTToken,
) -> models.Client:
    """
    Expand a client from the provided JWT token.
    """
    release, device = resolve_client_relations(token)
    client, _ = models.Client.objects.get_or_create(
        client_id=token.client_id,
        iss=token.iss,
        defaults={"release": release, "device": device},
    )

    if not client.device and device:
        client.device = device
        client.save(update_fields=["device"])

    if not client.release and release:
        client.release = release
        client.save(update_fields=["release"])

    return client



# Generated by Django 5.2 on 2025-12-05 09:38

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("authentikate", "0003_app_release_client_release"),
    ]

    operations = [
        migrations.CreateModel(
            name="Device",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("device_id", models.CharField(max_length=2000, unique=True)),
            ],
        ),
        migrations.AddField(
            model_name="client",
            name="device",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="authentikate.device",
            ),
        ),
    ]



import strawberry
from typing import Awaitable, Callable, Any, Optional, List, Union
from graphql import GraphQLError
from strawberry.extensions.field_extension import AsyncExtensionResolver
from strawberry.schema_directive import Location
from kante.types import Info
from strawberry.extensions import FieldExtension
from strawberry.schema_directive import Location
from strawberry.types.field import StrawberryField
from authentikate.base_models import JWTToken


@strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
class Auth:
    """A directive to enforce authentication and authorization on fields."""

    required_scopes: Optional[List[str]] = strawberry.directive_field(
        name="required_scopes", default=None
    )
    required_roles: Optional[List[str]] = strawberry.directive_field(
        name="required_roles", default=None
    )


class AuthExtension(FieldExtension):
    """A directive to enforce authentication and authorization on fields."""

    def __init__(
        self,
        scopes: Optional[List[str]] | str = None,
        roles: Optional[List[str]] = None,
        any_role_of: Optional[List[str]] = None,
        any_scope_of: Optional[List[str]] = None,
    ) -> None:
        """Initialize the AuthExtension with optional scopes and roles."""
        if isinstance(scopes, str):
            scopes = [scopes]
        if roles and isinstance(roles, str):
            roles = [roles]

        self.scopes: Optional[List[str]] = scopes
        self.roles: Optional[List[str]] = roles
        self.any_role_of: Optional[List[str]] = any_role_of
        self.any_scope_of: Optional[List[str]] = any_scope_of

    def apply(self, field: StrawberryField) -> None:
        """Apply the Auth directive to the field.

        Args:
            field (StrawberryField): The authentication field to which the directive will be applied.
        """
        assert (
            not field.is_subscription
        ), "Auth directive cannot be applied to subscriptions use AuthSubscribeExtension instead."
        field.directives.append(
            Auth(required_scopes=self.scopes, required_roles=self.roles)
        )

    def resolve(
        self, next_: Callable[..., Any], source: Any, info: Info, **kwargs
    ) -> Any:
        """Resolve the field with authentication checks."""
        if not info.context.request.user:
            raise GraphQLError("Authentication required")

        try:
            token: JWTToken = info.context.request.get_extension("token")

            if self.scopes and not token.has_scopes(self.scopes):
                raise GraphQLError(
                    f"User does not have the required scopes: {self.scopes}"
                )

            if self.any_scope_of and not token.has_any_role(self.any_scope_of):
                raise GraphQLError(
                    f"User does not have any of of the required scopes: {self.any_scope_of}"
                )

            if self.roles and not token.has_roles(self.roles):
                raise GraphQLError(
                    f"User does not have the required roles: {', '.join(self.roles)}"
                )

            if self.any_role_of and not token.has_any_role(self.any_role_of):
                raise GraphQLError(
                    f"User does not have any of the required roles: {', '.join(self.any_role_of)}"
                )

        except KeyError:
            raise GraphQLError("Token not found in request context")

        return next_(source, info, **kwargs)

    async def resolve_async(
        self,
        next_: Callable[..., Awaitable[Any]],
        source: Any,
        info: Info,
        **kwargs: Any,
    ) -> Any:
        """Resolve the field with authentication checks."""
        if not info.context.request.user:
            raise GraphQLError("Authentication required")

        try:
            token: JWTToken = info.context.request.get_extension("token")

            if self.scopes and not token.has_scopes(self.scopes):
                raise GraphQLError(
                    f"User does not have the required scopes: {self.scopes}"
                )

            if self.roles and not token.has_roles(self.roles):
                raise GraphQLError(
                    f"User does not have the required roles: {', '.join(self.roles)}"
                )

        except KeyError:
            raise GraphQLError("Token not found in request context")

        return await next_(source, info, **kwargs)


class AuthSubscribeExtension(FieldExtension):

    def __init__(
        self,
        scopes: Optional[List[str]] | str = None,
        roles: Optional[List[str]] = None,
    ) -> None:
        """Initialize the AuthExtension with optional scopes and roles."""
        if isinstance(scopes, str):
            scopes = [scopes]
        if roles and isinstance(roles, str):
            roles = [roles]

        self.scopes: Optional[List[str]] = scopes
        self.roles: Optional[List[str]] = roles

    def apply(self, field: StrawberryField) -> None:
        """Apply the Auth directive to the field.

        Args:
            field (StrawberryField): The authentication field to which the directive will be applied.
        """
        assert (
            field.is_subscription
        ), "AuthSubscribeExtension can only be applied to subscription fields."
        field.directives.append(
            Auth(required_scopes=self.scopes, required_roles=self.roles)
        )

    def resolve(
        self, next_: Callable[..., Any], source: Any, info: Info, **kwargs
    ) -> Any:
        """Resolve the field with authentication checks."""
        if not info.context.request.user:
            raise GraphQLError("Authentication required")

        try:
            token: JWTToken = info.context.request.get_extension("token")

            if self.scopes and not token.has_scopes(self.scopes):
                raise GraphQLError(
                    f"User does not have the required scopes: {self.scopes}"
                )

            if self.roles and not token.has_roles(self.roles):
                raise GraphQLError(
                    f"User does not have the required roles: {', '.join(self.roles)}"
                )

        except KeyError:
            raise GraphQLError("Token not found in request context")

        return next_(source, info, **kwargs)

    async def resolve_async(
        self,
        next_: Callable[..., Awaitable[Any]],
        source: Any,
        info: Info,
        **kwargs: Any,
    ) -> Any:
        """Resolve the field with authentication checks."""
        if not info.context.request.user:
            raise GraphQLError("Authentication required")

        try:
            token: JWTToken = info.context.request.get_extension("token")

            if self.scopes and not token.has_scopes(self.scopes):
                raise GraphQLError(
                    f"User does not have the required scopes: {self.scopes}"
                )

            if self.roles and not token.has_roles(self.roles):
                raise GraphQLError(
                    f"User does not have the required roles: {', '.join(self.roles)}"
                )

        except KeyError:
            raise GraphQLError("Token not found in request context")

        # this is a workaround for the fact that strawberry does not support async resolvers for subscriptions
        return next_(source, info, **kwargs)


all_directives = [Auth]



from authentikate import models
import kante
import strawberry


@kante.django_type(models.Device)
class Device:
    """This is the devicetype"""

    id: strawberry.ID
    device_id: str


@kante.django_type(models.App)
class App:
    """This is the apptype"""

    id: strawberry.ID
    identifier: str


@kante.django_type(models.Release)
class Release:
    """This is the release type"""

    id: strawberry.ID
    app: App
    version: str


@kante.django_type(models.Organization)
class Organization:
    """This is the organization type"""

    id: strawberry.ID
    slug: str


@kante.django_type(models.User)
class User:
    """This is the user type"""

    id: strawberry.ID
    sub: str
    preferred_username: str
    active_organization: Organization | None = None


@kante.django_type(models.Client)
class Client:
    """This is the client type"""

    id: strawberry.ID
    release: Release | None = None
    client_id: str
    name: str



# Generated by Django 5.2 on 2025-09-25 11:37

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("authentikate", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Membership",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("blocked", models.BooleanField(default=False)),
                ("roles", models.JSONField(default=list)),
                (
                    "organization",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="memberships",
                        to="authentikate.organization",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="memberships",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "unique_together": {("user", "organization")},
            },
        ),
    ]
